# -*- coding: utf-8 -*-

"""This module provides access to the process definition REST api of Camunda."""

from __future__ import annotations
import datetime as dt
import typing
import dataclasses

import pycamunda.variable
import pycamunda.processinst
import pycamunda.batch
import pycamunda.incident
import pycamunda.instruction
import pycamunda.base
from pycamunda.request import QueryParameter, PathParameter, BodyParameter, BodyParameterContainer

URL_SUFFIX = '/process-definition'


__all__ = [
    'GetActivityInstanceStats', 'GetProcessDiagram', 'Count', 'GetList', 'GetProcessInstanceStats',
    'GetXML', 'Get', 'StartInstance', 'Activate', 'Suspend', 'UpdateHistoryTimeToLive', 'Delete',
    'RestartProcessInstance'
]


@dataclasses.dataclass
class ProcessDefinition:
    """Data class of process definition as returned by the REST api of Camunda."""
    id_: str
    key: str
    category: str
    description: str
    name: str
    version: int
    resource: str
    deployment_id: str
    diagram: str
    suspended: bool
    tenant_id: str
    version_tag: str
    history_time_to_live: int
    startable_in_tasklist: bool

    @classmethod
    def load(cls, data: typing.Mapping[str, typing.Any]) -> ProcessDefinition:
        return cls(
            id_=data['id'],
            key=data['key'],
            category=data['category'],
            description=data['description'],
            name=data['name'],
            version=data['version'],
            resource=data['resource'],
            deployment_id=data['deploymentId'],
            diagram=data['diagram'],
            suspended=data['suspended'],
            tenant_id=data['tenantId'],
            version_tag=data['versionTag'],
            history_time_to_live=data['historyTimeToLive'],
            startable_in_tasklist=data['startableInTasklist']
        )


@dataclasses.dataclass
class ActivityStats:
    """Data class of activity statistics as returned by the REST api of Camunda."""
    id_: str
    instances: int
    failed_jobs: int
    incidents: typing.Iterable[pycamunda.incident.IncidentTypeCount]

    @classmethod
    def load(cls, data: typing.Mapping[str, typing.Any]) -> ActivityStats:
        return cls(
            id_=data['id'],
            instances=data['instances'],
            failed_jobs=data['failedJobs'],
            incidents=tuple(
                pycamunda.incident.IncidentTypeCount.load(incident_data)
                for incident_data in data['incidents']
            )
        )


@dataclasses.dataclass
class ProcessInstanceStats:
    """Data class of process instance statistics as returned by the REST api of Camunda."""
    id_: str
    instances: int
    failed_jobs: int
    definition: ProcessDefinition
    incidents: typing.Iterable[pycamunda.incident.IncidentTypeCount]

    @classmethod
    def load(cls, data: typing.Mapping[str, typing.Any]) -> ProcessInstanceStats:
        return cls(
            id_=data['id'],
            instances=data['instances'],
            failed_jobs=data['failedJobs'],
            definition=ProcessDefinition.load(data['definition']),
            incidents=tuple(
                pycamunda.incident.IncidentTypeCount.load(incident_data)
                for incident_data in data['incidents']
            )
        )


class GetActivityInstanceStats(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')
    key = PathParameter('key')
    tenant_id = PathParameter('tenant-id')

    failed_jobs = QueryParameter('failedJobs')
    incidents = QueryParameter('incidents', provide=pycamunda.base.value_is_true)
    incidents_for_type = QueryParameter('incidentsForType')

    def __init__(
        self,
        url: str,
        id_: str = None,
        key: str = None,
        tenant_id: str = None,
        failed_jobs: bool = None,
        incidents: bool = False,
        incidents_for_type: typing.Union[str, pycamunda.incident.IncidentType] = None
    ):
        """Get runtime statistics for a process definition. Does not include historic data.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the process definition.
        :param key: Key of the process definition.
        :param tenant_id: Id of the tenant the process definition belongs to.
        :param failed_jobs: Whether the number of failed jobs should be included.
        :param incidents: Whether to include the number of incidents.
        :param incidents_for_type: Include only incidents of a specific type.
        """
        if incidents and incidents_for_type is not None:
            raise pycamunda.PyCamundaException(
                'Either \'incidents\' or \'incidents_for_type\' can be provided, not both.'
            )
        super().__init__(url=url + URL_SUFFIX + '/{path}/statistics')
        self.id_ = id_
        self.key = key
        self.tenant_id = tenant_id
        self.failed_jobs = failed_jobs
        self.incidents = incidents
        self.incident_type = None
        if incidents_for_type is not None:
            self.incidents_for_type = pycamunda.incident.IncidentType(incidents_for_type)

    @property
    def url(self):
        if self.id_ is not None:
            return self._url.format(path=self.id_)
        if self.tenant_id is not None:
            return self._url.format(path=f'key/{self.key}/tenant-id/{self.tenant_id}')
        return self._url.format(path=f'key/{self.key}')

    def __call__(self, *args, **kwargs) -> typing.Tuple[ActivityStats]:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        return tuple(ActivityStats.load(activity_json) for activity_json in response.json())


class GetProcessDiagram(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')
    key = PathParameter('key')
    tenant_id = PathParameter('tenant-id')

    def __init__(self, url: str, id_: str = None, key: str = None, tenant_id: str = None):
        """Get the diagram of a process definition.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the process definition.
        :param key: Key of the process definition.
        """
        super().__init__(url=url + URL_SUFFIX + '/{path}/diagram')
        self.id_ = id_
        self.key = key
        self.tenant_id = tenant_id

    @property
    def url(self):
        if self.id_ is not None:
            return self._url.format(path=self.id_)
        if self.tenant_id is not None:
            return self._url.format(path=f'key/{self.key}/tenant-id/{self.tenant_id}')
        return self._url.format(path=f'key/{self.key}')

    def __call__(self, *args, **kwargs):
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        return response.content


class Count(pycamunda.base.CamundaRequest):

    id_ = QueryParameter('processDefinitionId')
    id_in = QueryParameter('processDefinitionIdIn')
    name = QueryParameter('name')
    name_like = QueryParameter('nameLike')
    deployment_id = QueryParameter('deploymentId')
    key = QueryParameter('key')
    key_in = QueryParameter('keysIn')
    key_like = QueryParameter('keyLike')
    category = QueryParameter('category')
    category_like = QueryParameter('categoryLike')
    version = QueryParameter('version')
    latest_version = QueryParameter('latestVersion', provide=pycamunda.base.value_is_true)
    resource_name = QueryParameter('resourceName')
    resource_name_like = QueryParameter('resourceNameLike')
    startable_by = QueryParameter('startableBy')
    active = QueryParameter('active', provide=pycamunda.base.value_is_true)
    suspended = QueryParameter('suspended', provide=pycamunda.base.value_is_true)
    incident_id = QueryParameter('incidentId')
    incident_type = QueryParameter('incidentType')
    incident_message = QueryParameter('incidentMessage')
    incident_message_like = QueryParameter('incidentMessageLike')
    tenant_id_in = QueryParameter('tenantIdIn')
    without_tenant_id = QueryParameter('withoutTenantId', provide=pycamunda.base.value_is_true)
    include_without_tenant_id = QueryParameter(
        'includeProcessDefinitionsWithoutTenantId',
        provide=pycamunda.base.value_is_true
    )
    version_tag = QueryParameter('versionTag')
    version_tag_like = QueryParameter('versionTagLike')
    without_version_tag = QueryParameter('withoutVersionTag')
    startable_in_tasklist = QueryParameter('startableInTasklist')
    not_startable_in_tasklist = QueryParameter('notStartableInTasklist')
    startable_permission_check = QueryParameter(
        'startablePermissionCheck',
        mapping={True: 'true', False: 'false'}
    )

    def __init__(
        self,
        url: str,
        id_: str = None,
        id_in: typing.Iterable[str] = None,
        name: str = None,
        name_like: str = None,
        deployment_id: str = None,
        key: str = None,
        key_in: typing.Iterable[str] = None,
        key_like: str = None,
        category: str = None,
        category_like: str = None,
        version: int = None,
        latest_version: bool = False,
        resource_name: str = None,
        resource_name_like: str = None,
        startable_by: str = None,
        active: bool = False,
        suspended: bool = False,
        incident_id: str = None,
        incident_type: typing.Union[str, pycamunda.incident.IncidentType] = None,
        incident_message: str = None,
        incident_message_like: str = None,
        tenant_id_in: typing.Iterable[str] = None,
        without_tenant_id: bool = False,
        include_without_tenant_id: bool = False,
        version_tag: str = None,
        version_tag_like: str = None,
        without_version_tag: bool = None,
        startable_in_tasklist: bool = None,
        startable_permission_check: bool = None,
        not_startable_in_tasklist: bool = None,
    ):
        """Count process definitions.

        :param url: Camunda Rest engine URL.
        :param id_: Filter by id.
        :param id_in: Filter whether the id is one of multiple ones.
        :param name: Filter by name.
        :param name_like: Filter by a substring of the name.
        :param deployment_id: Filter by deployment id.
        :param key: Key of the process definition.
        :param key_in: Filter whether the key is one of multiple ones.
        :param key_like: Filter by a substring of the key.
        :param category: Filter by category.
        :param category_like: Filter by a substring of the category.
        :param version: Filter by version.
        :param latest_version: Whether to include only the latest versions.
        :param resource_name: Filter by resource name.
        :param resource_name_like: Filter by a substring of the resource name.
        :param startable_by: Filter by user names that are allowed to start an instance of the
                             process definition.
        :param active: Whether to include only active process definitions.
        :param suspended: Whether to include only suspended process definitions.
        :param incident_id: Filter by incident id.
        :param incident_type: Filter by incident type.
        :param incident_message: Filter by the incident message.
        :param incident_message_like: Filter by a substring the incident message.
        :param tenant_id_in: Filter whether the tenant id is one of multiple ones.
        :param without_tenant_id: Whether to include only process definitions that belong to no
                                  tenant.
        :param include_without_tenant_id: Whether to include process definitions that belong to no
                                          tenant.
        :param version_tag: Filter by the version tag.
        :param version_tag_like: Filter by a substring of the version tag.
        :param without_version_tag: Whether to include only process definition without a version
                                    tag.
        :param startable_in_tasklist: Filter by process definition that are startable in tasklist.
        :param not_startable_in_tasklist: Filter by process definitions that are not startable in
                                          tasklist.
        :param startable_permission_check: Filter by process definitions that the user is allowed to
                                           start.
        """
        super().__init__(url=url + URL_SUFFIX + '/count')
        self.id_ = id_
        self.id_in = id_in
        self.name = name
        self.name_like = name_like
        self.deployment_id = deployment_id
        self.key = key
        self.key_like = key_like
        self.key_in = key_in
        self.category = category
        self.category_like = category_like
        self.version = version
        self.latest_version = latest_version
        self.resource_name = resource_name
        self.resource_name_like = resource_name_like
        self.startable_by = startable_by
        self.active = active
        self.suspended = suspended
        self.incident_id = incident_id
        self.incident_type = None
        if incident_type is not None:
            self.incident_type = pycamunda.incident.IncidentType(incident_type)
        self.incident_message = incident_message
        self.incident_message_like = incident_message_like
        self.tenant_id_in = tenant_id_in
        self.without_tenant_id = without_tenant_id
        self.include_without_tenant_id = include_without_tenant_id
        self.version_tag = version_tag
        self.version_tag_like = version_tag_like
        self.without_version_tag = without_version_tag
        self.startable_in_tasklist = startable_in_tasklist
        self.not_startable_in_tasklist = not_startable_in_tasklist
        self.startable_permission_check = startable_permission_check

    def __call__(self, *args, **kwargs) -> int:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        return response.json()['count']


class GetList(pycamunda.base.CamundaRequest):

    id_ = QueryParameter('processDefinitionId')
    id_in = QueryParameter('processDefinitionIdIn')
    name = QueryParameter('name')
    name_like = QueryParameter('nameLike')
    deployment_id = QueryParameter('deploymentId')
    key = QueryParameter('key')
    key_in = QueryParameter('keysIn')
    key_like = QueryParameter('keyLike')
    category = QueryParameter('category')
    category_like = QueryParameter('categoryLike')
    version = QueryParameter('version')
    latest_version = QueryParameter('latestVersion', provide=pycamunda.base.value_is_true)
    resource_name = QueryParameter('resourceName')
    resource_name_like = QueryParameter('resourceNameLike')
    startable_by = QueryParameter('startableBy')
    active = QueryParameter('active', provide=pycamunda.base.value_is_true)
    suspended = QueryParameter('suspended', provide=pycamunda.base.value_is_true)
    incident_id = QueryParameter('incidentId')
    incident_type = QueryParameter('incidentType')
    incident_message = QueryParameter('incidentMessage')
    incident_message_like = QueryParameter('incidentMessageLike')
    tenant_id_in = QueryParameter('tenantIdIn')
    without_tenant_id = QueryParameter('withoutTenantId', provide=pycamunda.base.value_is_true)
    include_without_tenant_id = QueryParameter(
        'includeProcessDefinitionsWithoutTenantId',
        provide = pycamunda.base.value_is_true
    )
    version_tag = QueryParameter('versionTag')
    version_tag_like = QueryParameter('versionTagLike')
    without_version_tag = QueryParameter('withoutVersionTag')
    startable_in_tasklist = QueryParameter('startableInTasklist')
    not_startable_in_tasklist = QueryParameter('notStartableInTasklist')
    startable_permission_check = QueryParameter(
        'startablePermissionCheck',
        mapping={True: 'true', False: 'false'}
    )
    sort_by = QueryParameter(
        'sortBy',
        mapping={
            'category': 'category',
            'key': 'key', 'id_': 'id',
            'name': 'name',
            'version': 'version',
            'deployment_id': 'deploymentId',
            'tenant_id': 'tenantId',
            'version_tag': 'versionTag'
        }
    )
    ascending = QueryParameter(
        'sortOrder',
        mapping={True: 'asc', False: 'desc'},
        provide=lambda self, obj, obj_type: vars(obj).get('sort_by', None) is not None
    )
    first_result = QueryParameter('firstResult')
    max_results = QueryParameter('maxResults')

    def __init__(
        self,
        url: str,
        id_: str = None,
        id_in: typing.Iterable[str] = None,
        name: str = None,
        name_like: str = None,
        deployment_id: str = None,
        key: str = None,
        key_in: typing.Iterable[str] = None,
        key_like: str = None,
        category: str = None,
        category_like: str = None,
        version: str = None,
        latest_version: bool = False,
        resource_name: str = None,
        resource_name_like: str = None,
        startable_by: str = None,
        active: bool = False,
        suspended: bool = False,
        incident_id: str = None,
        incident_type: typing.Union[str, pycamunda.incident.IncidentType] = None,
        incident_message: str = None,
        incident_message_like: str = None,
        tenant_id_in: typing.Iterable[str] = None,
        without_tenant_id: bool = False,
        include_without_tenant_id: bool = False,
        version_tag: str = None,
        version_tag_like: str = None,
        without_version_tag: bool = None,
        startable_in_tasklist: bool = None,
        startable_permission_check: bool = None,
        not_startable_in_tasklist: bool = None,
        sort_by: str = None,
        ascending: bool = True,
        first_result: int = None,
        max_results: int = None
    ):
        """Query for a list of process definitions using a list of parameters. The size of the
        result set can be retrieved by using the Get List Count method.

        :param url: Camunda Rest engine URL.
        :param id_: Filter by id.
        :param id_in: Filter whether the id is one of multiple ones.
        :param name: Filter by name.
        :param name_like: Filter by a substring of the name.
        :param deployment_id: Filter by deployment id.
        :param key: Key of the process definition.
        :param key_in: Filter whether the key is one of multiple ones.
        :param key_like: Filter by a substring of the key.
        :param category: Filter by category.
        :param category_like: Filter by a substring of the category.
        :param version: Filter by version.
        :param latest_version: Whether to include only the latest versions.
        :param resource_name: Filter by resource name.
        :param resource_name_like: Filter by a substring of the resource name.
        :param startable_by: Filter by user names that are allowed to start an instance of the
                             process definition.
        :param active: Whether to include only active process definitions.
        :param suspended: Whether to include only suspended process definitions.
        :param incident_id: Filter by incident id.
        :param incident_type: Filter by incident type.
        :param incident_message: Filter by the incident message.
        :param incident_message_like: Filter by a substring the incident message.
        :param tenant_id_in: Filter whether the tenant id is one of multiple ones.
        :param without_tenant_id: Whether to include only process definitions that belong to no
                                  tenant.
        :param include_without_tenant_id: Whether to include process definitions that belong to no
                                          tenant.
        :param version_tag: Filter by the version tag.
        :param version_tag_like: Filter by a substring of the version tag.
        :param without_version_tag: Whether to include only process definition without a version
                                    tag.
        :param startable_in_tasklist: Filter by process definition that are startable in tasklist.
        :param not_startable_in_tasklist: Filter by process definitions that are not startable in
                                          tasklist.
        :param startable_permission_check: Filter by process definitions that the user is allowed to
                                           start.
        :param sort_by: Sort the results by 'category', 'key', 'id_', 'name', 'version',
                        'deployment_id', 'tenant_id' or `version_tag`. Sorting by 'version_tag' is
                        string-based.
        :param ascending: Sort order.
        :param first_result: Pagination of results. Index of the first result to return.
        :param max_results: Pagination of results. Maximum number of results to return.
        """
        super().__init__(url=url + URL_SUFFIX)
        self.id_ = id_
        self.id_in = id_in
        self.name = name
        self.name_like = name_like
        self.deployment_id = deployment_id
        self.key = key
        self.key_like = key_like
        self.key_in = key_in
        self.category = category
        self.category_like = category_like
        self.version = version
        self.latest_version = latest_version
        self.resource_name = resource_name
        self.resource_name_like = resource_name_like
        self.startable_by = startable_by
        self.active = active
        self.suspended = suspended
        self.incident_id = incident_id
        self.incident_type = None
        if incident_type is not None:
            self.incident_type = pycamunda.incident.IncidentType(incident_type)
        self.incident_message = incident_message
        self.incident_message_like = incident_message_like
        self.tenant_id_in = tenant_id_in
        self.without_tenant_id = without_tenant_id
        self.include_without_tenant_id = include_without_tenant_id
        self.version_tag = version_tag
        self.version_tag_like = version_tag_like
        self.without_version_tag = without_version_tag
        self.startable_in_tasklist = startable_in_tasklist
        self.not_startable_in_tasklist = not_startable_in_tasklist
        self.startable_permission_check = startable_permission_check
        self.sort_by = sort_by
        self.ascending = ascending
        self.first_result = first_result
        self.max_results = max_results

    def __call__(self, *args, **kwargs) -> typing.Tuple[ProcessDefinition]:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        return tuple(ProcessDefinition.load(definition_json) for definition_json in response.json())


class GetProcessInstanceStats(pycamunda.base.CamundaRequest):

    failed_jobs = QueryParameter('failedJobs')
    incidents = QueryParameter('incidents', provide=pycamunda.base.value_is_true)
    root_incidents = QueryParameter('rootIncidents', provide=pycamunda.base.value_is_true)
    incidents_for_type = QueryParameter('incidentsForType')

    def __init__(
        self,
        url: str,
        failed_jobs: bool = False,
        incidents: bool = False,
        root_incidents: bool = False,
        incidents_for_type: typing.Union[str, pycamunda.incident.IncidentType] = None
    ):
        """Get runtime statistics grouped by process definition. Does not include historic data.

        :param url: Camunda Rest engine URL.
        :param failed_jobs: Whether the number of failed jobs should be included.
        :param incidents: Whether to include the number of incidents.
        :param root_incidents: Whether to include the corresponding number of root incidents for
                               each incident type.
        :param incidents_for_type: Include only incidents of a specific type.
        """
        if sum((incidents, root_incidents, incidents_for_type is not None)) > 1:
            raise pycamunda.PyCamundaException(
                'Either \'incidents\', \'root_incidents\' or \'incidents_for_type\' can be '
                'provided, not multiple of them.'
            )
        super().__init__(url=url + URL_SUFFIX + '/statistics')
        self.failed_jobs = failed_jobs
        self.incidents = incidents
        self.root_incidents = root_incidents
        self.incidents_for_type = None
        if incidents_for_type is not None:
            self.incidents_for_type = pycamunda.incident.IncidentType(incidents_for_type)

    def __call__(self, *args, **kwargs) -> typing.Tuple[ProcessInstanceStats]:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        return tuple(ProcessInstanceStats.load(stats_json) for stats_json in response.json())


class GetXML(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')
    key = PathParameter('key')
    tenant_id = PathParameter('tenant-id')

    def __init__(self, url: str, id_: str = None, key: str = None, tenant_id: str = None):
        """Get the BPMN xml diagram of a process definition.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the process definition.
        :param key: Key of the process definition.
        :param tenant_id: Id of the tenant the process definition belongs to.
        """
        super().__init__(url=url + URL_SUFFIX + '/{path}/xml')
        self.id_ = id_
        self.key = key
        self.tenant_id = tenant_id

    @property
    def url(self):
        if self.id_ is not None:
            return self._url.format(path=self.id_)
        if self.tenant_id is not None:
            return self._url.format(path=f'key/{self.key}/tenant-id/{self.tenant_id}')
        return self._url.format(path=f'key/{self.key}')

    def __call__(self, *args, **kwargs) -> str:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        return response.json()['bpmn20Xml']


class Get(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')
    key = PathParameter('key')
    tenant_id = PathParameter('tenant-id')

    def __init__(self, url: str, id_: str = None, key: str = None, tenant_id: str = None):
        """Get a process definition.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the process definition.
        :param key: Key of the process definition.
        :param tenant_id: Id of the tenant the process definition belongs to.

        """
        super().__init__(url=url + URL_SUFFIX + '/{path}')
        self.id_ = id_
        self.key = key
        self.tenant_id = tenant_id

    @property
    def url(self):
        if self.id_ is not None:
            return self._url.format(path=self.id_)
        if self.tenant_id is not None:
            return self._url.format(path=f'key/{self.key}/tenant-id/{self.tenant_id}')
        return self._url.format(path=f'key/{self.key}')

    def __call__(self, *args, **kwargs) -> ProcessDefinition:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        return ProcessDefinition.load(response.json())


class StartInstance(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')
    key = PathParameter('key')
    tenant_id = PathParameter('tenant-id')

    variables = BodyParameter('variables')
    business_key = BodyParameter('businessKey')
    case_instance_key = BodyParameter('caseInstanceId')
    start_instructions = BodyParameter('startInstructions')
    skip_custom_listeners = BodyParameter('skipCustomListeners')
    skip_io_mappings = BodyParameter('skipIoMappings')
    with_variables_in_return = BodyParameter('withVariablesInReturn')

    def __init__(
        self,
        url: str,
        id_: str = None,
        key: str = None,
        tenant_id: str = None,
        business_key: str = None,
        case_instance_id: str = None,
        skip_custom_listeners: bool = False,
        skip_io_mappings: bool = False,
        with_variables_in_return: bool = False
    ):
        """Start a process instance of a specific process definition.

        The process definition can be chosen by providing either the id or the key of the process
        definition. If chosen by key the latest version of that process definition is used. In case
        the key of the process definition is provided, the tenant id can be provided aswell.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the process definition.
        :param key: Key of the process definition.
        :param tenant_id: Id of the tenant the process definition belongs to.
        :param business_key: The business key to initialize the process instance with.
        :param case_instance_id: The case instance id to initialize the process instance with.
        :param skip_custom_listeners: Whether to skip custom listeners and notify only builtin ones.
        :param skip_io_mappings: Whether to skip input/output mappings.
        :param with_variables_in_return: Whether the variable that were used by the process instance
                                         during execution should be returned.
        """
        if id_ is not None and key is not None:
            raise pycamunda.PyCamundaException('Either `id_ or `key` can be provided, not both.')
        if tenant_id is not None and key is None:
            raise pycamunda.PyCamundaException(
                'If `tenant_id is provided `key also has to be provided.'
            )
        super().__init__(url=url + URL_SUFFIX + '/{path}/start')
        self.id_ = id_
        self.key = key
        self.tenant_id = tenant_id
        self.business_key = business_key
        self.case_instance_key = case_instance_id
        self.skip_custom_listeners = skip_custom_listeners
        self.skip_io_mappings = skip_io_mappings
        self.with_variables_in_return = with_variables_in_return

        self.variables = {}
        self.start_instructions = []

    @property
    def url(self):
        if self.id_ is not None:
            return self._url.format(path=self.id_)
        if self.tenant_id is not None:
            return self._url.format(path=f'key/{self.key}/tenant-id/{self.tenant_id}')
        return self._url.format(path=f'key/{self.key}')

    def add_variable(
        self, name: str, value: typing.Any, type_: str = None, value_info: str = None
    ) -> None:
        """Add a variable to initialize the process instance with.

        :param name: Name of the variable.
        :param value: Value of the variable.
        :param type_: Value type of the variable.
        :param value_info: Additional information regarding the value type.
        """
        self.variables[name] = {'value': value, 'type': type_, 'valueInfo': value_info}

    def _add_start_instruction(
        self,
        type_: typing.Union[str, pycamunda.instruction.InstructionType],
        activity_id: str = None,
        transition_id: str = None,
        variables: typing.Mapping[str, pycamunda.variable.Variable] = None
    ) -> None:
        """Add an instruction that specifies at which activities the process instance is started.

        :param type_: Type of the instruction. Possible values are
                          - startBeforeActivity
                          - startAfterActivity
                          - startTransition
        :param activity_id: Id of the activity in case `type_` is `startBeforeActivity` or
                            `startAfterActivity.
        :param transition_id: Id of the sequence flow to start.
        :param variables: Mapping from names to the corresponding variables.
        :return:
        """
        instruction = {'type': pycamunda.instruction.InstructionType(type_).value}
        if activity_id is not None:
            instruction['activityId'] = activity_id
        if transition_id is not None:
            instruction['transitionId'] = transition_id
        if variables is not None:
            instruction['variables'] = {
                name: dataclasses.asdict(var) for name, var in variables.items()
            }

        self.start_instructions.append(instruction)

    def add_start_before_activity_instruction(
        self,
        id_: str,
        variables: typing.Mapping[str, pycamunda.variable.Variable] = None
    ) -> None:
        """Add an instruction to start execution before a given activity is entered.

        :param id_: Id of the activity.
        :param variables: Mapping from names to the corresponding variables.
        """
        self._add_start_instruction(
            type_=pycamunda.instruction.InstructionType.start_before_activity,
            activity_id=id_,
            variables=variables
        )

    def add_start_after_activity_instruction(
        self,
        id_: str,
        variables: typing.Mapping[str, pycamunda.variable.Variable] = None
    ) -> None:
        """Add an instruction to start execution at the single outgoing sequence flow of an
        activity.

        :param id_: Id of the activity.
        :param variables: Mapping from names to the corresponding variables.
        """
        self._add_start_instruction(
            type_=pycamunda.instruction.InstructionType.start_after_activity,
            activity_id=id_,
            variables=variables
        )

    def add_start_transition_instruction(
        self,
        id_: str,
        variables: typing.Mapping[str, pycamunda.variable.Variable] = None
    ) -> None:
        """Add an instruction to start execution at the single outgoing sequence flow of an
        activity.

        :param id_: Id of the sequence flow to start.
        :param variables: Mapping from names to the corresponding variables.
        """
        self._add_start_instruction(
            type_=pycamunda.instruction.InstructionType.start_transition,
            transition_id=id_,
            variables=variables
        )

    def __call__(self, *args, **kwargs) -> pycamunda.processinst.ProcessInstance:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.POST, *args, **kwargs)

        return pycamunda.processinst.ProcessInstance.load(response.json())


class _ActivateSuspend(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')
    key = PathParameter('key')
    tenant_id = PathParameter('tenant-id')

    suspended = BodyParameter('suspended')
    include_process_instances = BodyParameter('includeProcessInstances')
    execution_datetime = BodyParameter('executionDate')

    def __init__(
        self,
        url: str,
        suspended: bool,
        id_: str = None,
        key: str = None,
        tenant_id: str = None,
        include_process_instances: bool = None,
        execution_datetime: dt.datetime = None
    ):
        """Activate or Suspend a process definition.

        :param url: Camunda Rest engine URL.
        :param suspended: Whether to suspend or activate the process definition.
        :param id_: Id of the process definition.
        :param key: Key of the process definition.
        :param tenant_id: Id of the tenant the process definition belongs to.
        :param include_process_instances: Whether to cascade the action to process instances.
        :param execution_datetime: When to execute the action. If 'None' the action is immediately.
        """
        super().__init__(url=url + URL_SUFFIX + '/{path}/suspended')
        self.suspended = suspended
        self.id_ = id_
        self.key = key
        self.tenant_id = tenant_id
        self.include_process_instances = include_process_instances
        self.execution_datetime = execution_datetime

    @property
    def url(self):
        if self.id_ is not None:
            return self._url.format(path=self.id_)
        if self.tenant_id is not None:
            return self._url.format(path=f'key/{self.key}/tenant-id/{self.tenant_id}')
        return self._url.format(path=f'key/{self.key}')

    def __call__(self, *args, **kwargs) -> None:
        """Send the request."""
        super().__call__(pycamunda.base.RequestMethod.PUT, *args, **kwargs)


class Activate(_ActivateSuspend):

    def __init__(
        self,
        url: str,
        id_: str = None,
        key: str = None,
        tenant_id: str = None,
        include_process_instances: bool = None,
        execution_datetime: dt.datetime = None
    ):
        """Activate a process definition.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the process definition.
        :param key: Key of the process definition.
        :param tenant_id: Id of the tenant the process definition belongs to.
        :param include_process_instances: Whether to cascade the action to process instances.
        :param execution_datetime: When to execute the action. If 'None' the action is immediately.
        """
        super().__init__(
            url=url,
            suspended=False,
            id_=id_,
            key=key,
            tenant_id=tenant_id,
            include_process_instances=include_process_instances,
            execution_datetime=execution_datetime
        )


class Suspend(_ActivateSuspend):

    def __init__(
            self,
            url: str,
            id_: str = None,
            key: str = None,
            tenant_id: str = None,
            include_process_instances: bool = None,
            execution_datetime: dt.datetime = None
    ):
        """Suspend a process definition.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the process definition.
        :param key: Key of the process definition.
        :param tenant_id: Id of the tenant the process definition belongs to.
        :param include_process_instances: Whether to cascade the action to process instances.
        :param execution_datetime: When to execute the action. If 'None' the action is immediately.
        """
        super().__init__(
            url=url,
            suspended=True,
            id_=id_,
            key=key,
            tenant_id=tenant_id,
            include_process_instances=include_process_instances,
            execution_datetime=execution_datetime
        )


class UpdateHistoryTimeToLive(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')
    key = PathParameter('key')
    tenant_id = PathParameter('tenant-id')

    history_time_to_live = BodyParameter(
        'historyTimeToLive',
        validate=lambda val: val is None or val >= 0
    )

    def __init__(
        self,
        url: str,
        history_time_to_live: int,
        id_: str = None,
        key: str = None,
        tenant_id: str = None
    ):
        """Update the history time to live of a process definition.

        :param url: Camunda Rest engine URL.
        :param history_time_to_live: New history time to live. Can be set to 'None'.
        :param id_: Id of the process definition.
        :param key: Key of the process definition.
        :param tenant_id: Id of the tenant the process definition belongs to.
        """
        super().__init__(url=url + URL_SUFFIX + '/{path}/history-time-to-live')
        self.history_time_to_live = history_time_to_live
        self.id_ = id_
        self.key = key
        self.tenant_id = tenant_id

    @property
    def url(self):
        if self.id_ is not None:
            return self._url.format(path=self.id_)
        if self.tenant_id is not None:
            return self._url.format(path=f'key/{self.key}/tenant-id/{self.tenant_id}')
        return self._url.format(path=f'key/{self.key}')

    def __call__(self, *args, **kwargs) -> None:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.PUT, *args, **kwargs)


class Delete(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')
    key = PathParameter('key')
    tenant_id = PathParameter('tenant-id')

    cascade = QueryParameter('cascade')
    skip_custom_listeners = QueryParameter('skipCustomListeners')
    skip_io_mappings = QueryParameter('skipIoMappings')

    def __init__(
        self,
        url: str,
        id_: str = None,
        key: str = None,
        tenant_id: str = None,
        cascade: bool = False,
        skip_custom_listeners: bool = False,
        skip_io_mappings: bool = False
    ):
        """Delete a process definition.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the process definition.
        :param key: Key of the process definition.
        :param tenant_id: Id of the tenant the process definition belongs to.
        :param cascade: Whether to cascade the deletion to process instances of the definition.
        :param skip_custom_listeners: Whether to skip custom listeners and notify only builtin ones.
        :param skip_io_mappings: Whether to skip input/output mappings.
        """
        super().__init__(url=url + URL_SUFFIX + '/{path}')
        self.id_ = id_
        self.key = key
        self.tenant_id = tenant_id
        self.cascade = cascade
        self.skip_custom_listeners = skip_custom_listeners
        self.skip_io_mappings = skip_io_mappings

    @property
    def url(self):
        if self.id_ is not None:
            return self._url.format(path=self.id_)
        if self.tenant_id is not None:
            return self._url.format(path=f'key/{self.key}/tenant-id/{self.tenant_id}')
        return self._url.format(path=f'key/{self.key}')

    def __call__(self, *args, **kwargs) -> None:
        """Send the request."""
        super().__call__(pycamunda.base.RequestMethod.DELETE, *args, **kwargs)


class RestartProcessInstance(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')
    process_instance_ids = BodyParameter('processInstanceIds')
    historic_process_instance_query = BodyParameterContainer('historicProcessInstanceQuery')  # TODO create method to add this parameter
    skip_custom_listeners = BodyParameter('skipCustomListeners')
    skip_io_mappings = BodyParameter('skipIoMappings')
    initial_variables = BodyParameter('initialVariables')
    without_business_key = BodyParameter('withoutBusinessKey')
    instructions = BodyParameter('instructions')

    def __init__(
        self,
        url: str,
        id_: str,
        process_instance_ids: typing.Iterable[str],
        async_: bool = False,
        skip_custom_listeners: bool = False,
        skip_io_mappings: bool = False,
        initial_variables: bool = True,
        without_business_key: bool = False
    ):
        """Restart process instances of a specific process definition.

        :param url: Camunda Rest engine url.
        :param id_: Id of the process definition.
        :param process_instance_ids: Ids of the process instances to restart.
        :param async_: Whether to restart the processes asynchronously.
        :param skip_custom_listeners: Whether to skip custom listeners and notify only builtin ones.
        :param skip_io_mappings: Whether to skip input/output mappings.
        :param initial_variables: Whether to set the initial set of variables.
        :param without_business_key: Whether not to add the business key of the process instance.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}/restart')
        self.id_ = id_
        self.process_instance_ids = process_instance_ids
        self.async_ = async_
        self.skip_custom_listeners = skip_custom_listeners
        self.skip_io_mappings = skip_io_mappings
        self.initial_variables = initial_variables
        self.without_business_key = without_business_key

        self.instructions = []

    @property
    def url(self) -> str:
        return super().url + ('-async' if self.async_ else '')

    def _add_instruction(
        self,
        type_: typing.Union[str, pycamunda.instruction.InstructionType],
        activity_id: str = None,
        transition_id: str = None
    ) -> None:
        """Add an instruction that specifies at which activities the process instance is started.

        :param type_: Type of the instruction. Possible values are
                          - startBeforeActivity
                          - startAfterActivity
                          - startTransition
        :param activity_id: Id of the activity in case `type_` is `startBeforeActivity` or
                            `startAfterActivity.
        :param transition_id: Id of the sequence flow to start.
        """
        instruction = {'type': pycamunda.instruction.InstructionType(type_).value}
        if activity_id is not None:
            instruction['activityId'] = activity_id
        if transition_id is not None:
            instruction['transitionId'] = transition_id

        self.instructions.append(instruction)

    def add_before_activity_instruction(self, id_: str) -> None:
        """Add an instruction to start execution before a given activity is entered.

        :param id_: Id of the activity.
        """
        self._add_instruction(
            type_=pycamunda.instruction.InstructionType.start_before_activity,
            activity_id=id_
        )

    def add_after_activity_instruction(self, id_: str) -> None:
        """Add an instruction to start execution at the single outgoing sequence flow of an
        activity.

        :param id_: Id of the activity.
        """
        self._add_instruction(
            type_=pycamunda.instruction.InstructionType.start_after_activity,
            activity_id=id_
        )

    def add_transition_instruction(self, id_: str) -> None:
        """Add an instruction to start execution at the single outgoing sequence flow of an
        activity.

        :param id_: Id of the sequence flow to start.
        """
        self._add_instruction(
            type_=pycamunda.instruction.InstructionType.start_transition,
            transition_id=id_
        )

    def __call__(self, *args, **kwargs) -> typing.Optional[pycamunda.batch.Batch]:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.POST, *args, **kwargs)

        if self.async_:
            return pycamunda.batch.Batch.load(response.json())



# -*- coding: utf-8 -*-

"""This module provides access to the external task REST api of Camunda."""

from __future__ import annotations
import datetime as dt
import dataclasses
import typing

import requests

import pycamunda
import pycamunda.base
import pycamunda.variable
import pycamunda.batch
from pycamunda.request import QueryParameter, PathParameter, BodyParameter

URL_SUFFIX = '/external-task'


__all__ = [
    'GetList', 'Count', 'FetchAndLock', 'Complete', 'HandleBPMNError', 'HandleFailure', 'Unlock',
    'ExtendLock', 'SetPriority', 'SetRetries', 'SetRetriesAsync', 'SetRetriesSync'
]


@dataclasses.dataclass
class ExternalTask:
    """Data class of external task as returned by the REST api of Camunda."""
    activity_id: str
    activity_instance_id: str
    error_message: str
    error_details: str
    execution_id: str
    id_: str
    process_definition_id: str
    process_definition_key: str
    process_instance_id: str
    tenant_id: str
    retries: int
    worker_id: str
    priority: str
    topic_name: str
    lock_expiration_time: dt.datetime = None
    suspended: bool = None
    business_key: str = None
    variables: typing.Dict[str, typing.Dict[str, typing.Any]] = None

    @classmethod
    def load(cls, data: typing.Mapping[str, typing.Any]) -> ExternalTask:
        external_task = cls(
            activity_id=data['activityId'],
            activity_instance_id=data['activityInstanceId'],
            error_message=data['errorMessage'],
            error_details=data['errorDetails'],
            execution_id=data['executionId'],
            id_=data['id'],
            process_definition_id=data['processDefinitionId'],
            process_definition_key=data['processDefinitionKey'],
            process_instance_id=data['processInstanceId'],
            tenant_id=data['tenantId'],
            retries=data['retries'],
            worker_id=data['workerId'],
            priority=data['priority'],
            topic_name=data['topicName']
        )
        if data['lockExpirationTime'] is not None:
            external_task.lock_expiration_time = pycamunda.base.from_isoformat(
                data['lockExpirationTime']
            )
        try:
            external_task.suspended = data['suspended']
        except KeyError:
            pass
        try:
            external_task.business_key = data['businessKey']
        except KeyError:
            pass
        try:
            variables = data['variables']
        except KeyError:
            pass
        else:
            external_task.variables = {
                var_name: pycamunda.variable.Variable(
                    type_=var['type'], value=var['value'], value_info=var['valueInfo']
                )
                for var_name, var in variables.items()
            }
        return external_task


class Get(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')

    def __init__(self, url: str, id_: str, request_error_details: bool = True):
        """Query for an external task.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the external task.
        :param request_error_details: Whether to request error details for tasks. Requires an
                                      additional request.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}')
        self.id_ = id_
        self.request_error_details = request_error_details

    def __call__(self, *args, **kwargs) -> ExternalTask:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)
        external_task = ExternalTask.load(response.json())

        if self.request_error_details:
            if external_task.error_details is None:
                try:
                    response = requests.get(self.url + '/errorDetails')
                except requests.exceptions.RequestException:
                    raise pycamunda.PyCamundaException()
                if not response:
                    pycamunda.base._raise_for_status(response)
                external_task.error_details = response.text

        return external_task


class GetList(pycamunda.base.CamundaRequest):

    id_ = QueryParameter('externalTaskId')
    topic_name = QueryParameter('topicName')
    worker_id = QueryParameter('workerId')
    locked = QueryParameter('locked', provide=pycamunda.base.value_is_true)
    not_locked = QueryParameter('notLocked', provide=pycamunda.base.value_is_true)
    with_retries_left = QueryParameter('withRetriesLeft', provide=pycamunda.base.value_is_true)
    no_retries_left = QueryParameter('noRetriesLeft', provide=pycamunda.base.value_is_true)
    lock_expiration_after = QueryParameter('lockExpirationAfter')
    lock_expiration_before = QueryParameter('lockExpirationBefore')
    activity_id = QueryParameter('activityId')
    actitity_id_in = QueryParameter('activityIdIn')
    execution_id = QueryParameter('executionId')
    process_instance_id = QueryParameter('processInstanceId')
    process_definition_id = QueryParameter('processDefinitionId')
    tenant_id_in = QueryParameter('tenantIdIn')
    active = QueryParameter('active', provide=pycamunda.base.value_is_true)
    priority_higher_equals = QueryParameter('priorityHigherThanOrEquals')
    priority_lower_equals = QueryParameter('priorityLowerThanOrEquals')
    suspended = QueryParameter('suspended', provide=pycamunda.base.value_is_true)
    sort_by = QueryParameter(
        'sortBy',
        mapping={
         'id_': 'id',
         'lock_expiration_time': 'lockExpirationTime',
         'process_instance_id': 'processInstanceId',
         'process_definition_id': 'processDefinitionId',
         'tenant_id': 'tenantId',
         'task_priority': 'taskPriority'
        }
    )
    ascending = QueryParameter(
        'sortOrder',
        mapping={True: 'asc', False: 'desc'},
        provide=lambda self, obj, obj_type: vars(obj).get('sort_by', None) is not None
    )
    first_result = QueryParameter('firstResult')
    max_results = QueryParameter('maxResults')

    def __init__(
        self,
        url: str,
        id_: str = None,
        topic_name: str = None,
        worker_id: str = None,
        locked: bool = False,
        not_locked: bool = False,
        with_retries_left: bool = False,
        no_retries_left: bool = False,
        lock_expiration_after: dt.datetime = None,
        lock_expiration_before: dt.datetime = None,
        activity_id: str = None,
        activity_id_in: typing.Iterable[str] = None,
        execution_id: str = None,
        process_instance_id: str = None,
        process_definition_id: str = None,
        tenant_id_in: typing.Iterable[str] = None,
        active: bool = False,
        priority_higher_equals: int = None,
        priority_lower_equals: int = None,
        suspended: bool = False,
        sort_by: str = None,
        ascending: bool = True,
        first_result: int = None,
        max_results: int = None,
        request_error_details: bool = True
    ):
        """Query for a list of external tasks using a list of parameters. The size of the result set
        can be retrieved by using the Get Count request.

        :param url: Camunda Rest engine URL.
        :param id_: Filter by the id of the external task.
        :param topic_name: Filter by the topic name of the external task.
        :param worker_id: Filter by the id of the worker the task was locked by last.
        :param locked: Include only locked external tasks.
        :param not_locked: Include only unlocked tasks.
        :param with_retries_left: Include only external tasks that have retries left.
        :param no_retries_left: Include only external tasks that have no retries left.
        :param lock_expiration_after: Include only external tasks with a lock that expires after the
                                      provided date.
        :param lock_expiration_before: Include only external tasks with a lock that expires before
                                       the provided date.
        :param activity_id: Filter by activity id the external task is created for.
        :param activity_id_in: Filter whether activity id is one of multiple ones.
        :param execution_id: Filter by the execution id the external task belongs to.
        :param process_instance_id: Filter by the process instance id the external task belongs to.
        :param process_definition_id: Filter by the process definition id the external task belongs
                                      to.
        :param tenant_id_in: Filter whether the tenant id is one of multiple ones.
        :param active: Include only external tasks that are active.
        :param priority_higher_equals: Include only external tasks with a priority higher than or
                                       equals to the given value.
        :param priority_lower_equals: Include only external tasks with a priority lower than or
                                      equals to the given value.
        :param suspended: Include only external tasks that are suspended.
        :param sort_by: Sort the results by `id_`, `lock_expiration_time, `process_instance_id`,
                        `process_definition_key`, `tenant_id` or `task_priority`.
        :param ascending: Sort order.
        :param first_result: Pagination of results. Index of the first result to return.
        :param max_results: Pagination of results. Maximum number of results to return.
        :param request_error_details: Whether to request error details for tasks. Requires
                                      additional requests.
        """
        super().__init__(url=url + URL_SUFFIX)
        self.id_ = id_
        self.topic_name = topic_name
        self.worker_id = worker_id
        self.locked = locked
        self.not_locked = not_locked
        self.with_retries_left = with_retries_left
        self.no_retries_left = no_retries_left
        self.lock_expiration_after = lock_expiration_after
        self.lock_expiration_before = lock_expiration_before
        self.activity_id = activity_id
        self.actitity_id_in = activity_id_in
        self.execution_id = execution_id
        self.process_instance_id = process_instance_id
        self.process_definition_id = process_definition_id
        self.tenant_id_in = tenant_id_in
        self.active = active
        self.priority_higher_equals = priority_higher_equals
        self.priority_lower_equals = priority_lower_equals
        self.suspended = suspended
        self.sort_by = sort_by
        self.ascending = ascending
        self.first_result = first_result
        self.max_results = max_results
        self.request_error_details = request_error_details

    def __call__(self, *args, **kwargs) -> typing.Tuple[ExternalTask]:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)
        external_tasks = tuple(ExternalTask.load(task_json) for task_json in response.json())

        if self.request_error_details:
            for external_task in external_tasks:
                if external_task.error_details is None:
                    try:
                        response = requests.get(self.url + f'/{external_task.id_}/errorDetails')
                    except requests.exceptions.RequestException:
                        raise pycamunda.PyCamundaException()
                    if not response:
                        pycamunda.base._raise_for_status(response)
                    external_task.error_details = response.text

        return external_tasks


class Count(GetList):

    def __init__(
        self,
        url: str,
        id_: str = None,
        topic_name: str = None,
        worker_id: str = None,
        locked: bool = False,
        not_locked: bool = False,
        with_retries_left: bool = False,
        no_retries_left: bool = False,
        lock_expiration_after: dt.datetime = None,
        lock_expiration_before: dt.datetime = None,
        activity_id: str = None,
        activity_id_in: typing.Iterable[str] = None,
        execution_id: str = None,
        process_instance_id: str = None,
        process_definition_id: str = None,
        tenant_id_in: typing.Iterable[str] = None,
        active: bool = False,
        priority_higher_equals: int = None,
        priority_lower_equals: int = None,
        suspended: bool = False,
        sort_by: str = None,
        ascending: bool = True,
        first_result: int = None,
        max_results: int = None
    ):
        """Get the size of the result returned by the Get List request.

        :param url: Camunda Rest engine URL.
        :param id_: Filter by the id of the external task.
        :param topic_name: Filter by the topic name of the external task.
        :param worker_id: Filter by the id of the worker the task was locked by last.
        :param locked: Include only locked external tasks.
        :param not_locked: Include only unlocked tasks.
        :param with_retries_left: Include only external tasks that have retries left.
        :param no_retries_left: Include only external tasks that have no retries left.
        :param lock_expiration_after: Include only external tasks with a lock that expires after a
                                      date.
        :param lock_expiration_before: Include only external tasks with a lock that expires before a
                                       date.
        :param activity_id: Filter by activity id the external task is created for.
        :param activity_id_in: Filter whether activity id is one of multiple ones.
        :param execution_id: Filter by the execution id the external task belongs to.
        :param process_instance_id: Filter by the process instance id the external task belongs to.
        :param process_definition_id: Filter by the process definition id the external task belongs
                                      to.
        :param tenant_id_in: Filter whether the tenant id is one of multiple ones.
        :param active: Include only external tasks that are active.
        :param priority_higher_equals: Include only external tasks with a priority higher than or
                                       equals to the given value.
        :param priority_lower_equals: Include only external tasks with a priority lower than or
                                      equals to the given value.
        :param suspended: Include only external tasks that are suspended.
        :param sort_by: Sort the results by `id_`, `lock_expiration_time, `process_instance_id`,
                        `process_definition_key`, `tenant_id` or `task_priority`.
        :param ascending: Sort order.
        :param first_result: Pagination of results. Index of the first result to return.
        :param max_results: Pagination of results. Maximum number of results to return.
        """
        super().__init__(
            url=url,
            id_=id_,
            topic_name=topic_name,
            worker_id=worker_id,
            locked=locked,
            not_locked=not_locked,
            with_retries_left=with_retries_left,
            no_retries_left=no_retries_left,
            lock_expiration_after=lock_expiration_after,
            lock_expiration_before=lock_expiration_before,
            activity_id=activity_id,
            activity_id_in=activity_id_in,
            execution_id=execution_id,
            process_instance_id=process_instance_id,
            process_definition_id=process_definition_id,
            tenant_id_in=tenant_id_in,
            active=active,
            priority_higher_equals=priority_higher_equals,
            priority_lower_equals=priority_lower_equals,
            suspended=suspended,
            sort_by=sort_by,
            ascending=ascending,
            first_result=first_result,
            max_results=max_results
        )
        self._url += '/count'

    def query_parameters(self, apply: typing.Callable = ...):
        params = super().query_parameters(apply=apply)
        for key, value in params.items():
            if isinstance(value, dt.datetime):
                params[key] = pycamunda.base.isoformat(datetime_=value)
        return params

    def __call__(self, *args, **kwargs) -> int:
        """Send the request."""
        try:
            response = requests.request(
                method=pycamunda.base.RequestMethod.GET.value,
                url=self.url,
                params=self.query_parameters()
            )
        except requests.exceptions.RequestException:
            raise pycamunda.PyCamundaException()
        if not response:
            pycamunda.base._raise_for_status(response)

        return int(response.json()['count'])


class FetchAndLock(pycamunda.base.CamundaRequest):

    worker_id = BodyParameter('workerId')
    max_tasks = BodyParameter('maxTasks')
    use_priority = BodyParameter('usePriority')
    async_response_timeout = BodyParameter('asyncResponseTimeout')
    topics = BodyParameter('topics')

    def __init__(
        self,
        url: str,
        worker_id: str,
        max_tasks: int,
        async_response_timeout: int = None,
        use_priority: bool = False
    ):
        """Fetch and lock external tasks for a specific worker. Only external tasks with topics that
        are added to this request are fetched.

        :param url: Camunda Rest engine URL.
        :param worker_id: Id of the worker the external tasks are locked for.
        :param max_tasks: Maximum number of tasks to fetch.
        :param async_response_timeout: Long polling timeout in milliseconds. Cannot be set larger
                                       than 1800000 (30 minutes).
        :param use_priority: Whether the tasks should be fetched based on their priority.
        """
        super().__init__(url=url + URL_SUFFIX + '/fetchAndLock')
        self.worker_id = worker_id
        self.max_tasks = max_tasks
        self.async_response_timeout = async_response_timeout
        self.use_priority = use_priority
        self.topics = []

    def add_topic(
        self,
        name: str,
        lock_duration: int,
        variables: typing.Iterable[str] = None,
        deserialize_values: bool = False
    ) -> None:
        """Add a topic to this request.

        :param name: Name of the topic.
        :param lock_duration: Duration to lock the fetched external tasks for in milliseconds.
        :param variables: Variables to request from the process instance the external task belongs
                          to. If set to `None` all variables are requested.
        :param deserialize_values: Whether serializable variable values are deserialized on server
                                   side.
        """
        topic = {
            'topicName': name,
            'lockDuration': lock_duration,
            'deserializeValues': deserialize_values
        }
        if variables is not None:
            topic['variables'] = variables
        self.topics.append(topic)

    def __call__(self, *args, **kwargs) -> typing.Tuple[ExternalTask]:
        """Send the request"""
        response = super().__call__(pycamunda.base.RequestMethod.POST, *args, **kwargs)

        return tuple(ExternalTask.load(task_json) for task_json in response.json())


class Complete(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')
    worker_id = BodyParameter('workerId')
    variables = BodyParameter('variables')
    local_variables = BodyParameter('localVariables')

    def __init__(self, url: str, id_: str, worker_id: str):
        """Complete an external task that is locked for a worker.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the external task.
        :param worker_id: Id of the worker the external tasks was locked for.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}/complete')
        self.id_ = id_
        self.worker_id = worker_id
        self.variables = {}
        self.local_variables = {}

    def add_variable(
        self, name: str, value: str, type_: str = None, value_info: typing.Mapping = None
    ) -> None:
        """Add a variable to send to the Camunda process instance.

        :param name: Name of the variable.
        :param value: Value of the variable.
        :param type_: Value type of the variable.
        :param value_info: Additional information regarding the value type.
        """
        self.variables[name] = {'value': value, 'type': type_, 'valueInfo': value_info}

    def add_local_variable(
        self, name: str, value: str, type_: str = None, value_info: typing.Mapping = None
    ) -> None:
        """Add a local variable to send to Camunda. Local variables are set only in the scope of an
        external task.

        :param name: Name of the variable.
        :param value: Value of the variable.
        :param type_: Value type of the variable.
        :param value_info: Additional information regarding the value type.
        """
        self.local_variables[name] = {'value': value, 'type': type_, 'valueInfo': value_info}

    def __call__(self, *args, **kwargs) -> None:
        """Send the request"""
        super().__call__(pycamunda.base.RequestMethod.POST, *args, **kwargs)


class HandleBPMNError(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')
    worker_id = BodyParameter('workerId')
    error_code = BodyParameter('errorCode')
    error_message = BodyParameter('errorMessage')
    variables = BodyParameter('variables')

    def __init__(
        self,
        url: str,
        id_: str,
        worker_id: str,
        error_code: str,
        error_message: str = None
    ):
        """Report a business error for a running external task.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the external task.
        :param worker_id: Id of the worker that locked the external task.
        :param error_code: Error code that identifies the predefined error.
        :param error_message: Error message that describes the error.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}/bpmnError')
        self.id_ = id_
        self.worker_id = worker_id
        self.error_code = error_code
        self.error_message = error_message
        self.variables = {}

    def add_variable(
        self, name: str, value: str, type_: str = None, value_info: typing.Mapping = None
    ) -> None:
        """Add a variable to send to the Camunda process instance.

        :param name: Name of the variable.
        :param value: Value of the variable.
        :param type_: Value type of the variable.
        :param value_info: Additional information regarding the value type.
        """
        self.variables[name] = {'value': value, 'type': type_, 'valueInfo': value_info}

    def __call__(self, *args, **kwargs) -> None:
        """Send the request"""
        super().__call__(pycamunda.base.RequestMethod.POST, *args, **kwargs)


class HandleFailure(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')
    worker_id = BodyParameter('workerId')
    error_message = BodyParameter('errorMessage')
    error_details = BodyParameter('errorDetails')
    retries = BodyParameter('retries', validate=lambda val: val >= 0)
    retry_timeout = BodyParameter('retryTimeout', validate=lambda val: val >= 0)

    def __init__(
        self,
        url: str,
        id_: str,
        worker_id: str,
        error_message: str,
        error_details: str,
        retries: int,
        retry_timeout: int
    ):
        """Report a failure to execute a running external task.

        A number of retries and a timeout until the external task can be tried can be specified.
        If retries are set to 0, the external task cannot be fetched anymore and an incident is
        created. The message of the incident is set to the value of `error_message`.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the external task.
        :param worker_id: Id of the worker that locked the external task.
        :param error_message: Error message that describes the reason of the failure.
        :param error_details: Error description.
        :param retries: How often the external task can be retried.
        :param retry_timeout: Timeout in milliseconds until the external task becomes available
        again for fetching.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}/failure')
        self.id_ = id_
        self.worker_id = worker_id
        self.error_message = error_message
        self.error_details = error_details
        self.retries = retries
        self.retry_timeout = retry_timeout

    def __call__(self, *args, **kwargs) -> None:
        """Send the request"""
        super().__call__(pycamunda.base.RequestMethod.POST, *args, **kwargs)


class Unlock(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')

    def __init__(self, url: str, id_: str):
        """Unlock an external task.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the external task.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}/unlock')
        self.id_ = id_

    def __call__(self, *args, **kwargs) -> None:
        """Send the request"""
        super().__call__(pycamunda.base.RequestMethod.POST, *args, **kwargs)


class ExtendLock(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')
    new_duration = BodyParameter('newDuration')
    worker_id = BodyParameter('workerId')

    def __init__(self, url: str, id_: str, new_duration: int, worker_id: str):
        """Unlock an external task.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the external task.
        :param new_duration: New duration in milliseconds how long the external task wants to be
                             locked.
        :param worker_id: Id of the worker that locked this external task.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}/extendLock')
        self.id_ = id_
        self.new_duration = new_duration
        self.worker_id = worker_id

    def __call__(self, *args, **kwargs) -> None:
        """Send the request"""
        super().__call__(pycamunda.base.RequestMethod.POST, *args, **kwargs)


class SetPriority(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')
    priority = BodyParameter('priority')

    def __init__(self, url: str, id_: str, priority: int):
        """Sets the priority of an external task.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the external task.
        :param priority: New priority of the external task.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}/priority')
        self.id_ = id_
        self.priority = priority

    def __call__(self, *args, **kwargs) -> None:
        """Send the request"""
        super().__call__(pycamunda.base.RequestMethod.PUT, *args, **kwargs)


class SetRetries(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')
    retries = BodyParameter('retries', validate=lambda val: val >= 0)

    def __init__(self, url: str, id_: str, retries: int):
        """Sets the number of retries of an external task.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the external task.
        :param retries: New number of retries of the external task.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}/retries')
        self.id_ = id_
        self.retries = retries

    def __call__(self, *args, **kwargs) -> None:
        """Send the request"""
        super().__call__(pycamunda.base.RequestMethod.PUT, *args, **kwargs)


class SetRetriesAsync(pycamunda.base.CamundaRequest):

    retries = BodyParameter('retries', validate=lambda val: val >= 0)
    external_task_ids = BodyParameter('externalTaskIds')
    process_instance_ids = BodyParameter('processInstanceIds')
    external_task_query = BodyParameter('externalTaskQuery')
    process_instance_query = BodyParameter('processInstanceQuery')
    historic_process_instance_query = BodyParameter('historicProcessInstanceQuery')

    def __init__(self, url: str, retries: int, external_task_ids: typing.Iterable[str]):
        """Sets the number of retries of external tasks asynchronously.

        :param url: Camunda Rest engine URL.
        :param retries: New number of retries of the external tasks.
        :param external_task_ids: Ids of the external tasks.
        """
        super().__init__(url=url + URL_SUFFIX + '/retries-async')
        self.retries = retries
        self.external_task_ids = external_task_ids
        self.process_instance_ids = None  # TODO
        self.external_task_query = None  # TODO
        self.process_instance_query = None  # TODO
        self.historic_process_instance_query = None  # TODO

    def __call__(self, *args, **kwargs) -> pycamunda.batch.Batch:
        """Send the request"""
        response = super().__call__(pycamunda.base.RequestMethod.POST, *args, **kwargs)

        return pycamunda.batch.Batch.load(response.json())


class SetRetriesSync(pycamunda.base.CamundaRequest):

    retries = BodyParameter('retries', validate=lambda val: val >= 0)
    external_task_ids = BodyParameter('externalTaskIds')
    process_instance_ids = BodyParameter('processInstanceIds')
    external_task_query = BodyParameter('externalTaskQuery')
    process_instance_query = BodyParameter('processInstanceQuery')
    historic_process_instance_query = BodyParameter('historicProcessInstanceQuery')

    def __init__(self, url: str, retries: int, external_task_ids: typing.Iterable[str]):
        """Sets the number of retries of external tasks synchronously.

        :param url: Camunda Rest engine URL.
        :param retries: New number of retries of the external tasks.
        :param external_task_ids: Ids of the external tasks.
        """
        super().__init__(url=url + URL_SUFFIX + '/retries')
        self.retries = retries
        self.external_task_ids = external_task_ids
        self.process_instance_ids = None  # TODO
        self.external_task_query = None  # TODO
        self.process_instance_query = None  # TODO
        self.historic_process_instance_query = None  # TODO

    def __call__(self, *args, **kwargs) -> None:
        """Send the request"""
        super().__call__(pycamunda.base.RequestMethod.PUT, *args, **kwargs)



# -*- coding: utf-8 -*-
from __future__ import annotations


class PyCamundaException(Exception):
    """Base class for all PyCamunda exceptions."""


class NoSuccess(PyCamundaException):
    """Exception that is raised when the tried action is not successful."""
    http_code = None


class BadRequest(NoSuccess):
    """Exception that is raised when the tried action was invalid."""
    http_code = 400


class Unauthorized(NoSuccess):
    """Exception that is raised when the user is not authorized."""
    http_code = 401


class Forbidden(NoSuccess):
    """Exception that is raised when the tried action was valid but not permitted."""
    http_code = 403


class NotFound(NoSuccess):
    """Exception that is raised when the requested resource was not found."""
    http_code = 404


class InternalServerError(PyCamundaException):
    """Exception that is raised when there occurred an error on Camunda server side."""
    http_code = 500
